/*
	
	PWRBTN driver 1.3
	
	Install directory      : /boot/home/config/add-ons/kernel/drivers/bin
	Create Link file under : /boot/home/config/add-ons/kernel/drivers/dev/misc
	
	Copyright (C) 2003 momoziro
	
*/

#include "pwrbtn.h"

static pci_module_info *pci;
static int pmbase;
static pci_info info;

/*
	Intel
*/
void pm_intel_init(void)
{
	/* enable power management I/O space ( for PIIX4 )*/
	if (info.device_id == 0x7113) {
		write_pci_config(info.bus, info.device, info.function, 0x80, 1, 0x01);
	}
	
	/* get ACPI base address */
	pmbase = read_pci_config(info.bus, info.device, info.function, 0x40, 4);
	if (info.device_id == 0x7113) {
		pmbase &= 0xffc0;	// 1111 1111 1100 0000
	} else {
		pmbase &= 0xff80;	// 1111 1111 1000 0000
	}
	
	/* disable power button enable bit */
	(*pci->write_io_16) (pmbase + 2, 0);
}

bool pm_intel_pwrbtn_check(uint32 *pdata)
{
	bool pressed = false;
	
	*pdata = (*pci->read_io_16) (pmbase + 0);
	if (*pdata & 0x100) {	// check power button status
		// reset power button status bit
		(*pci->write_io_16) (pmbase + 0, 0x100);
		pressed = true;
	}
	
	return pressed;
}

/*
	SiS
*/
void pm_sis_init(void)
{
	/* get ACPI base address */
	pmbase = read_pci_config(info.bus, info.device, info.function, 0x68, 2);
	pmbase &= 0xffe0;	// 1111 1111 1110 0000
	
	/* disable power button enable bit */
	(*pci->write_io_16) (pmbase + 2, 0x0);
}

bool pm_sis_pwrbtn_check(uint32 *pdata)
{
	bool pressed = false;
	
	*pdata = (*pci->read_io_16) (pmbase + 0);
	if (*pdata & 0x100) {	// check power button status
		// reset power button status bit
		(*pci->write_io_16) (pmbase + 0, 0x100);
		pressed = true;
	}
	
	return pressed;
}

/*
	AMD
*/
void pm_amd_init(void)
{
	/* get ACPI base address */
	pmbase = read_pci_config(info.bus, info.device, info.function, 0x58, 2);
	pmbase &= 0xff00;
	
	/* disable power button enable bit */
	(*pci->write_io_16) (pmbase + 2, 0x0);
}

bool pm_amd_pwrbtn_check(uint32 *pdata)
{
	bool pressed = false;
	
	*pdata = (*pci->read_io_16) (pmbase + 0);
	if (*pdata & 0x100) {	// check power button status
		// reset power button status bit
		(*pci->write_io_16) (pmbase + 0, 0x100);
		pressed = true;
	}
	
	return pressed;
}
/*
	VIA
*/
void pm_via_init(void)
{
	/* get ACPI base address */
	switch(info.device_id) {
		case 0x3057:	// VT82C686
		case 0x8235:	// VT8231
			pmbase = read_pci_config(info.bus, info.device, info.function, 0x48, 4);
			break;
		case 0x3074:	// VT8233A, VT8233CE
		case 0x3177:	// VT8235
			pmbase = read_pci_config(info.bus, info.device, info.function, 0x88, 4);
			break;
	}
	pmbase &= 0xff00; // VIA datasheets state that I/O space is 256 bytes,
					  // and so I guess &= 0xff00 is okay.
	/* disable power button enable bit */
	(*pci->write_io_16) (pmbase + 2, 0x0);
}

bool pm_via_pwrbtn_check(uint32* pdata)
{
	bool pressed = false;
	
	*pdata = (*pci->read_io_16) (pmbase + 0);
	if(*pdata & 0x100) { // check power button status
		// reset power button status bit
		(*pci->write_io_16) (pmbase + 0, 0x100);
		pressed = true;
	}
	return pressed;
}


/* -----
	null-terminated array of device names supported by this driver
----- */

static const char *pwrbtn_name[] = {
	"misc/pwrbtn",
	NULL
};

/* -----
	search_device
----- */
status_t
search_device(void)
{
	int i;
	for (i = 0; (*pci->get_nth_pci_info)(i, &info) == B_OK; i++) {
//		dprintf("pwrbtn: vid=0x%x did=0x%x\n", info.vendor_id, info.device_id);
		switch(info.vendor_id) {
			case VID_INTEL: {	// Intel
				switch(info.device_id) {
					case 0x7113:	// 440LX/ZX/BX(82371AB PIIX4)
					case 0x2410:	// 810(82801AB)
					case 0x2420:	// 810(82801AB)
					case 0x2440:	// 815/850(ICH2)
					case 0x244c:	// 815/850(ICH2-M)
					case 0x248c:	// 830/845MP
					case 0x24c0:	// 845
					case 0x24cc:	// 855
					case 0x24d0:	// 865/875
						return B_OK;
				}
				break;
			}
			case VID_SIS: {	// SiS
				switch(info.device_id) {
					case 0x0630:	// SiS630
					case 0x0730:	// SiS730
						return B_OK;
				}
				break;
			}
			case VID_AMD: {	// AMD
				switch(info.device_id) {
					case 0x7413:	// AMD-766
					case 0x7443:	// AMD-768
					case 0x740b:	// AMD-756
						return B_OK;
				}
				break;
			}
			case VID_VIA: { // VIA
				switch(info.device_id) {
					case 0x3057:	// VT82C686
					case 0x8235:	// VT8231
					case 0x3074:	// VT8233A, VT8233CE
					case 0x3177:	// VT8235
						return B_OK;
				}
				break;
			}
		}
	}
	
	return ENODEV;
}

/* ----------
	init_hardware - called once the first time the driver is loaded
----- */
_EXPORT status_t
init_hardware (void)
{
	status_t err;
	
//	set_dprintf_enabled(true);
	dprintf ("pwrbtn: init_hardware()\n");
	if (get_module (B_PCI_MODULE_NAME, (module_info **)&pci)) return ENOSYS;
	err = search_device();
	if (err == B_OK) {
		/* found */
		dprintf("pwrbtn: vendor_id=0x%x device_id=0x%x\n", info.vendor_id, info.device_id);
	} else {
		dprintf("pwrbtn: device not found.\n");
	}
	put_module (B_PCI_MODULE_NAME);
	return err;
}


/* ----------
	init_driver - optional function - called every time the driver
	is loaded.
----- */
_EXPORT status_t
init_driver (void)
{
	status_t err;
	dprintf ("pwrbtn: init_driver()\n");
	if (get_module (B_PCI_MODULE_NAME, (module_info **)&pci)) return ENOSYS;
	err = search_device();
	if (err == B_OK) {
		/* found */
		switch (info.vendor_id) {
			case VID_INTEL:
				pm_intel_init();
				break;
			case VID_SIS:
				pm_sis_init();
				break;
			case VID_AMD:
				pm_amd_init();
				break;
			case VID_VIA:
				pm_via_init();
				break;
		}
		dprintf("pwrbtn: vendor_id=0x%x device_id=0x%x pmbase=0x%x\n", info.vendor_id, info.device_id, pmbase);
	} else {
		dprintf("pwrbtn: device not found.\n");
	}
	put_module (B_PCI_MODULE_NAME);
	return err;
}


/* ----------
	uninit_driver - optional function - called every time the driver
	is unloaded
----- */
_EXPORT void
uninit_driver (void)
{
}

/* ----------
	publish_devices - return a null-terminated array of devices
	supported by this driver.
----- */

_EXPORT const char**
publish_devices()
{
	return pwrbtn_name;
}

/* ----------
	open - handle open() calls
----- */

static status_t
pwrbtn_open (const char *name, uint32 flags, void** cookie)
{
	int i;
	const char *s;
	
	dprintf ("pwrbtn: open(%s)\n", name);
	for (i = 0; (s = pwrbtn_name[i]) != NULL; i++) {
		if (strcmp (name, s) == 0) break;
	}
	if (s == NULL) {
		dprintf ("pwrbtn: no such device\n");
		return ENODEV;
	}
	return B_OK;
}


/* ----------
	read - handle read() calls
----- */

static status_t
pwrbtn_read (void* cookie, off_t position, void *buf, size_t* num_bytes)
{
	return B_IO_ERROR;
}


/* ----------
	write - handle write() calls
----- */

static status_t
pwrbtn_write (void* cookie, off_t position, const void* buffer, size_t* num_bytes)
{
	return B_IO_ERROR;
}


/* ----------
	control - handle ioctl calls
----- */

static status_t
pwrbtn_control (void* cookie, uint32 op, void* arg, size_t len)
{
	struct pwrbtn_t *data = (struct pwrbtn_t *)arg;
	
	switch(info.vendor_id) {
		case VID_INTEL:
			if (pm_intel_pwrbtn_check(&data->pdata)) data->pwrbtn_pressed_count++;
			break;
		case VID_SIS:
			if (pm_sis_pwrbtn_check(&data->pdata)) data->pwrbtn_pressed_count++;
			break;
		case VID_AMD:
			if (pm_amd_pwrbtn_check(&data->pdata)) data->pwrbtn_pressed_count++;
			break;
		case VID_VIA:
			if (pm_via_pwrbtn_check(&data->pdata)) data->pwrbtn_pressed_count++;
			break;
		default: return B_ERROR;
	}
	
	return B_OK;
}


/* ----------
	close - handle close() calls
----- */

static status_t
pwrbtn_close (void* cookie)
{
	return B_OK;
}


/* -----
	free - called after the last device is closed, and after
	all i/o is complete.
----- */
static status_t
pwrbtn_free (void* cookie)
{
	return B_OK;
}


/* -----
	function pointers for the device hooks entry points
----- */

device_hooks pwrbtn_hooks = {
	pwrbtn_open, 			/* -> open entry point */
	pwrbtn_close, 			/* -> close entry point */
	pwrbtn_free,			/* -> free cookie */
	pwrbtn_control, 		/* -> control entry point */
	pwrbtn_read,			/* -> read entry point */
	pwrbtn_write			/* -> write entry point */
};

/* ----------
	find_device - return ptr to device hooks structure for a
	given device name
----- */

_EXPORT device_hooks*
find_device(const char* name)
{
	return &pwrbtn_hooks;
}
